Imports madrat.Minesweeper.Algorithms
Imports Microsoft.VisualBasic

''' <summary>
''' Class representing a Minesweeper game.
''' </summary>
Public Class MinesweeperGame

#Region " Private Types "

    Private Enum FieldValue
        Mined
        Valued
    End Enum

    <Flags()> Private Enum FieldState
        Hidden = 0
        Open = 1
        Pressed = 2
        Unknown = 4
        Flagged = 8
        WrongBomb = 16
        ClickedBomb = 32
    End Enum

    Private Enum GameState
        Started
        Running
        Paused
        Won
        Lost
    End Enum

    Private Enum MouseActionType
        MouseDown
        MouseUp
    End Enum

    <Flags()> Private Enum ButtonFlags
        None = 0
        Left = 1
        Right = 2
    End Enum

    Private Class Field
        Public Value As FieldValue
        Public State As FieldState
        Public Number As Integer
        Public DoGraphicalUpdate As Boolean = True

        Public Sub New()
        End Sub

        Public Sub New(ByVal value As FieldValue, ByVal state As FieldState, ByVal number As Integer)
            Me.Value = value
            Me.State = state
            Me.Number = number
        End Sub
    End Class

#End Region

#Region " Public Types "

    Public Enum FixedConfigurations
        Beginner
        Advanced
        Expert
    End Enum

    Public Event GameWon(ByVal sender As MinesweeperGame)
    Public Event GameLost(ByVal sender As MinesweeperGame)
    Public Event GameStarted(ByVal sender As MinesweeperGame)
    Public Event Clicked(ByVal sender As MinesweeperGame)
    Public Event Tick(ByVal sender As MinesweeperGame, ByVal e As TickEventArgs)
    Public Event RequestContainerResize(ByVal sender As MinesweeperGame, ByVal e As RequestContainerResizeEventArgs)

#End Region

#Region " Private Members "

    Private Const DefaultFieldSize As Integer = 9
    Private Const DefaultNumberOfMines As Integer = 10

    Private Shared _beginnerGame As New MinesweeperGame(DefaultFieldSize, DefaultFieldSize, DefaultNumberOfMines)
    Private Shared _advancedGame As New MinesweeperGame(16, 16, 40)
    Private Shared _expertGame As New MinesweeperGame(30, 16, 99)

    Private m_width As Integer = DefaultFieldSize
    Private m_height As Integer = DefaultFieldSize
    Private m_mines As Integer = DefaultNumberOfMines
    Private m_useMarker As Boolean = True
    Private m_field(DefaultFieldSize - 1, DefaultFieldSize - 1) As Field
    Private WithEvents m_uiDevice As PictureBox
    Private m_bitmap As Drawing.Bitmap
    Private m_elapsedTime As Integer
    Private m_running As Boolean = False
    Private WithEvents m_timer As Windows.Forms.Timer
    Private m_gameState As GameState
    Private m_lastButton As ButtonFlags = ButtonFlags.None
    Private m_activeFields As New List(Of Pair(Of Integer, Integer))
    Private m_freeFields As Integer
    Private m_minesLeft As Integer

#End Region

#Region " Public Interface "

#Region " c'tors "

    ''' <summary>
    ''' Creates a new class instance.
    ''' </summary>
    Public Sub New()

    End Sub

    ''' <summary>
    ''' Creates a new class instance by output device.
    ''' </summary>
    ''' <param name="graphicsDevice">An output device for the game field</param>
    ''' <param name="timer">A timer.</param>
    Public Sub New(ByVal graphicsDevice As PictureBox, ByVal timer As Windows.Forms.Timer)
        Me.m_uiDevice = graphicsDevice
        Me.m_timer = timer
    End Sub

    ''' <summary>
    ''' Creates a new class instance by game values.
    ''' </summary>
    ''' <param name="width">Game field width.</param>
    ''' <param name="height">Game field height.</param>
    ''' <param name="mines">Number of mines on field.</param>
    Public Sub New(ByVal width As Integer, ByVal height As Integer, ByVal mines As Integer)
        InitValues(width, height, mines)
    End Sub

    ''' <summary>
    ''' Creates a new class instance by output device and game values.
    ''' </summary>
    ''' <param name="graphicsDevice">An output device for the game field</param>
    ''' <param name="timer">A timer.</param>
    ''' <param name="width">Game field width.</param>
    ''' <param name="height">Game field height.</param>
    ''' <param name="mines">Number of mines on field.</param>
    Public Sub New( _
        ByVal graphicsDevice As PictureBox, _
        ByVal timer As Windows.Forms.Timer, _
        ByVal width As Integer, ByVal height As Integer, ByVal mines As Integer _
    )
        Me.New(graphicsDevice, timer)
        InitValues(width, height, mines)
    End Sub

#End Region

    ''' <summary>
    ''' Sets or returns the graphical user interface.
    ''' </summary>
    ''' <value><see cref="Windows.Forms.PictureBox">Windows.Forms.PictureBox</see></value>
    Public Property GraphicsDevice() As PictureBox
        Get
            Return Me.m_uiDevice
        End Get
        Set(ByVal value As PictureBox)
            Me.m_uiDevice = value
        End Set
    End Property

    ''' <summary>
    ''' Sets or returs the game timer.
    ''' </summary>
    ''' <value><see cref="Windows.Forms.Timer">Windows.Forms.Timer</see></value>
    Public Property Timer() As Windows.Forms.Timer
        Get
            Return Me.m_timer
        End Get
        Set(ByVal value As Windows.Forms.Timer)
            Me.m_timer = value
        End Set
    End Property

    ''' <summary>
    ''' Sets or returns whether the marker is used.
    ''' </summary>
    ''' <value><see cref="System.Boolean">Boolean</see></value>
    Public Property UseMarker() As Boolean
        Get
            Return Me.m_useMarker
        End Get
        Set(ByVal value As Boolean)
            Me.m_useMarker = value
        End Set
    End Property

    ''' <summary>
    ''' Returns the width of the field.
    ''' </summary>
    Public ReadOnly Property Width() As Integer
        Get
            Return Me.m_width
        End Get
    End Property

    ''' <summary>
    ''' Returns the height of the field.
    ''' </summary>
    Public ReadOnly Property Height() As Integer
        Get
            Return Me.m_height
        End Get
    End Property

    ''' <summary>
    ''' Returns the number of mines on the field.
    ''' </summary>
    Public ReadOnly Property Mines() As Integer
        Get
            Return Me.m_mines
        End Get
    End Property

    ''' <summary>
    ''' Returns the number of mines left (as the user marks them).
    ''' </summary>
    ''' <value><see cref="System.Integer">Integer</see></value>
    Public ReadOnly Property MinesLeft() As Integer
        Get
            Return Me.m_minesLeft
        End Get
    End Property

    ''' <summary>
    ''' Returns the elapsed playing time.
    ''' </summary>
    ''' <value><see cref="System.Integer">Integer</see></value>
    Public ReadOnly Property ElapsedTime() As Integer
        Get
            Return Me.m_elapsedTime
        End Get
    End Property

    ''' <summary>
    ''' Begins a new game with given settings.
    ''' </summary>
    Public Sub BeginGame()
        ReDim m_field(Me.m_width - 1, Me.m_height - 1)

        Dim newSize As New Drawing.Size(16 * Me.m_width + 2, 16 * Me.m_height + 1)
        RaiseEvent RequestContainerResize(Me, New RequestContainerResizeEventArgs(newSize))
        InitGraphics()

        For x As Integer = 0 To Me.m_width - 1
            For y As Integer = 0 To Me.m_height - 1
                Me.m_field(x, y) = New Field(FieldValue.Valued, FieldState.Hidden, 0)
            Next
        Next

        Me.m_activeFields.Clear()
        Me.m_freeFields = 0
        Me.m_elapsedTime = 0
        Me.m_minesLeft = Me.m_mines
        Me.m_timer.Interval = 1000

        Me.m_gameState = GameState.Started
        Me.m_running = True
        Redraw()
    End Sub

    ''' <summary>
    ''' Begins a new game with a preset configuration.
    ''' </summary>
    ''' <param name="presetConfig">A preset configuration.</param>
    Public Sub BeginGame(ByVal presetConfig As FixedConfigurations)
        Dim gameConfig As MinesweeperGame

        Select Case presetConfig
            Case FixedConfigurations.Beginner
                gameConfig = _beginnerGame
            Case FixedConfigurations.Advanced
                gameConfig = _advancedGame
            Case FixedConfigurations.Expert
                gameConfig = _expertGame
        End Select

        With gameConfig
            InitValues(.m_width, .m_height, .m_mines)
        End With

        BeginGame()
    End Sub

    ''' <summary>
    ''' Begins a game with custom parametres.
    ''' </summary>
    ''' <param name="width">Field width.</param>
    ''' <param name="height">Field height.</param>
    ''' <param name="mines">Number of mines.</param>
    Public Sub BeginGame(ByVal width As Integer, ByVal height As Integer, ByVal mines As Integer)
        InitValues(width, height, mines)
        BeginGame()
    End Sub

    ''' <summary>
    ''' Pauses and resumes the game.
    ''' </summary>
    Public Sub Pause()
        If Me.m_gameState = GameState.Running Then
            Me.m_timer.Stop()
            Me.m_gameState = GameState.Paused
            Redraw()
        ElseIf Me.m_gameState = GameState.Paused Then
            Me.m_timer.Start()
            Me.m_gameState = GameState.Running

            For Each field As Field In Me.m_field
                field.DoGraphicalUpdate = True
            Next

            Redraw()
        End If
    End Sub

#End Region

#Region " Private Implementation "

    Private Sub InitValues(ByVal width As Integer, ByVal height As Integer, ByVal mines As Integer)
        Me.m_width = width
        Me.m_height = height
        Me.m_mines = mines
    End Sub

    Private Sub m_timer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles m_timer.Tick
        Me.m_elapsedTime += 1
        RaiseEvent Tick(Me, New TickEventArgs(Me.m_elapsedTime))
    End Sub

    Private Sub m_uiDevice_MouseDown(ByVal sender As Object, ByVal e As MouseEventArgs) Handles m_uiDevice.MouseDown
        ReactOnInput(e, MouseActionType.MouseDown)
    End Sub

    Private Sub m_uiDevice_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles m_uiDevice.MouseMove
        If e.Button <> MouseButtons.None Then
            ReactOnInput(e, MouseActionType.MouseDown)
        End If
    End Sub

    Private Sub m_uiDevice_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles m_uiDevice.MouseUp
        ReactOnInput(e, MouseActionType.MouseUp)
    End Sub

    Private Sub m_uiDevice_Paint(ByVal sender As Object, ByVal e As PaintEventArgs) Handles m_uiDevice.Paint
        e.Graphics.DrawImage(Me.m_bitmap, 0, 0)
    End Sub

    Private Sub ReactOnInput(ByVal e As MouseEventArgs, ByVal action As MouseActionType)
        Dim doGraphicalRefresh As Boolean = False
        Static Dim lastField As Pair(Of Integer, Integer)

        ' preliminary operations '

        If Not Me.m_running Then Exit Sub

        Dim field As Pair(Of Integer, Integer) = FieldFromCursor(e)

        ' low-level event pre-processing '

        If e.Button = MouseButtons.Left Then
            If action = MouseActionType.MouseDown Then
                Me.m_lastButton = Me.m_lastButton Or ButtonFlags.Left
            End If
        ElseIf e.Button = MouseButtons.Right Then
            If action = MouseActionType.MouseDown Then
                Me.m_lastButton = Me.m_lastButton Or ButtonFlags.Right
            End If
        End If

        ' reaction on event '

        Select Case Me.m_gameState
            Case GameState.Started
                If Me.m_lastButton = ButtonFlags.Left Then
                    If action = MouseActionType.MouseDown Then
                        ' left button down '
                        If lastField IsNot Nothing Then RemoveActiveField(lastField)
                        lastField = field
                        Me.m_activeFields.Add(field)
                        doGraphicalRefresh = True
                    ElseIf action = MouseActionType.MouseUp Then
                        ' left button up '
                        ClearActiveFields()
                        lastField = Nothing
                        doGraphicalRefresh = True
                        StartNewGame(field)
                        DiscoverField(field)
                    End If
                End If
            Case GameState.Running
                If Me.m_lastButton = (ButtonFlags.Left Or ButtonFlags.Right) Then
                    If action = MouseActionType.MouseDown Then
                        ' both buttons down '
                        If lastField Is Nothing Then
                            MarkActiveSurroundingFields(field)
                        Else
                            ClearActiveFields()
                            MarkActiveSurroundingFields(field)
                        End If

                        lastField = field
                        doGraphicalRefresh = True
                    ElseIf action = MouseActionType.MouseUp Then
                        ' both buttons up '
                        ClearActiveFields()
                        DiscoverSurroundingFields(field)
                        doGraphicalRefresh = True
                    End If
                ElseIf Me.m_lastButton = ButtonFlags.Left Then
                    If action = MouseActionType.MouseDown Then
                        ' left button down '
                        If lastField IsNot Nothing Then RemoveActiveField(lastField)
                        lastField = field
                        Me.m_activeFields.Add(field)
                        doGraphicalRefresh = True
                    ElseIf action = MouseActionType.MouseUp Then
                        ' left button up '
                        ClearActiveFields()
                        lastField = Nothing
                        DiscoverField(field)
                        doGraphicalRefresh = True
                    End If
                ElseIf Me.m_lastButton = ButtonFlags.Right Then
                    If action = MouseActionType.MouseDown Then
                        ' right button down '
                    ElseIf action = MouseActionType.MouseUp Then
                        ' right button up '
                        ChangeFieldState(Me.m_field(field.V1, field.V2))
                        doGraphicalRefresh = True
                    End If
                End If
            Case GameState.Paused
            Case GameState.Lost
            Case GameState.Won
        End Select

        ' low-level event post-processing '

        If e.Button = MouseButtons.Left Then
            If action = MouseActionType.MouseUp Then
                Me.m_lastButton = Me.m_lastButton And (Not ButtonFlags.Left)
            End If
        ElseIf e.Button = MouseButtons.Right Then
            If action = MouseActionType.MouseUp Then
                Me.m_lastButton = Me.m_lastButton And (Not ButtonFlags.Right)
            End If
        End If

        If doGraphicalRefresh Then
            RaiseEvent Clicked(Me)
            Redraw()

            Select Case Me.m_gameState
                Case GameState.Won
                    RaiseEvent GameWon(Me)
                Case GameState.Lost
                    RaiseEvent GameLost(Me)
            End Select
        End If
    End Sub

    Private Sub RemoveActiveField(ByVal field As Pair(Of Integer, Integer))
        Me.m_field(field.V1, field.V2).DoGraphicalUpdate = True
        Me.m_activeFields.Remove(field)
    End Sub

    Private Sub ClearActiveFields()
        For Each field As Pair(Of Integer, Integer) In Me.m_activeFields
            Me.m_field(field.V1, field.V2).DoGraphicalUpdate = True
        Next

        Me.m_activeFields.Clear()
    End Sub

    Private Function FieldFromCursor(ByVal e As MouseEventArgs) As Pair(Of Integer, Integer)
        Return New Pair(Of Integer, Integer)( _
            (e.X + 2) * Me.m_width \ Me.m_uiDevice.Width, _
            (e.Y + 2) * Me.m_height \ Me.m_uiDevice.Height _
        )
    End Function

    Private Sub StartNewGame(ByVal fromField As Pair(Of Integer, Integer))
        Dim allFields As New List(Of Pair(Of Integer, Integer))

        For x As Integer = 0 To Me.m_width - 1
            For y As Integer = 0 To Me.m_height - 1
                If x <> fromField.V1 AndAlso y <> fromField.V2 Then
                    allFields.Add(New Pair(Of Integer, Integer)(x, y))
                End If
            Next
        Next

        Dim randomMines As Stack(Of Pair(Of Integer, Integer)) = Shuffle(allFields)

        For mine As Integer = 1 To Me.m_mines
            Dim pos As Pair(Of Integer, Integer) = randomMines.Pop()
            Me.m_field(pos.V1, pos.V2).Value = FieldValue.Mined

            For i As Integer = -1 To 1
                If pos.V1 + i >= 0 AndAlso pos.V1 + i < Me.m_width Then
                    For j As Integer = -1 To 1
                        If pos.V2 + j >= 0 AndAlso pos.V2 + j < Me.m_height Then
                            Me.m_field(pos.V1 + i, pos.V2 + j).Number += 1
                        End If
                    Next
                End If
            Next
        Next

        Me.m_timer.Start()
        Me.m_gameState = GameState.Running
        RaiseEvent GameStarted(Me)
    End Sub

    Private Sub DiscoverField(ByVal field As Pair(Of Integer, Integer))
        With Me.m_field(field.V1, field.V2)
            If (.State And FieldState.Open) = FieldState.Open Then Return
            If (.State And FieldState.Flagged) = FieldState.Flagged Then Return

            .State = .State Or FieldState.Open
            .DoGraphicalUpdate = True

            If .Value = FieldValue.Mined Then
                .State = .State Or FieldState.ClickedBomb
                GameIsLost()
                Return
            ElseIf .Number = 0 Then
                For i As Integer = -1 To 1
                    If field.V1 + i >= 0 AndAlso field.V1 + i < Me.m_width Then
                        For j As Integer = -1 To 1
                            If field.V2 + j >= 0 AndAlso field.V2 + j < Me.m_height Then
                                Dim newField As New Pair(Of Integer, Integer)(field.V1 + i, field.V2 + j)
                                If Not (Me.m_field(newField.V1, newField.V2).State And FieldState.Open) = FieldState.Open Then
                                    DiscoverField(newField)
                                End If
                            End If
                        Next
                    End If
                Next
            End If
        End With

        Me.m_freeFields += 1

        If Me.m_freeFields + Me.m_mines = Me.m_width * Me.m_height Then
            GameIsWon()
        End If
    End Sub

    Private Sub MarkActiveSurroundingFields(ByVal field As Pair(Of Integer, Integer))
        Me.m_activeFields.Add(field)

        With Me.m_field(field.V1, field.V2)
            For i As Integer = -1 To 1
                If field.V1 + i >= 0 AndAlso field.V1 + i < Me.m_width Then
                    For j As Integer = -1 To 1
                        If field.V2 + j >= 0 AndAlso field.V2 + j < Me.m_height Then
                            If Not ( _
                                (Me.m_field(field.V1 + i, field.V2 + j).State And FieldState.Open) = _
                                FieldState.Open OrElse _
                                (Me.m_field(field.V1 + i, field.V2 + j).State And FieldState.Flagged) = _
                                FieldState.Flagged _
                            ) Then
                                Me.m_activeFields.Add(New Pair(Of Integer, Integer)(field.V1 + i, field.V2 + j))
                            End If
                        End If
                    Next
                End If
            Next
        End With
    End Sub

    Private Sub DiscoverSurroundingFields(ByVal field As Pair(Of Integer, Integer))
        With Me.m_field(field.V1, field.V2)
            If Not (.State And FieldState.Open) = FieldState.Open Then Return

            For i As Integer = -1 To 1
                If field.V1 + i >= 0 AndAlso field.V1 + i < Me.m_width Then
                    For j As Integer = -1 To 1
                        If field.V2 + j >= 0 AndAlso field.V2 + j < Me.m_height Then
                            Dim newField As New Pair(Of Integer, Integer)(field.V1 + i, field.V2 + j)

                            If Not ( _
                                (Me.m_field(newField.V1, newField.V2).State And FieldState.Open) = _
                                FieldState.Open OrElse _
                                (Me.m_field(newField.V1, newField.V2).State And FieldState.Flagged) = _
                                FieldState.Flagged _
                            ) Then
                                If CountMinesAroundField(field) = Me.m_field(field.V1, field.V2).Number Then
                                    DiscoverField(newField)
                                End If
                            End If
                        End If
                    Next
                End If
            Next
        End With
    End Sub

    Private Function CountMinesAroundField(ByVal field As Pair(Of Integer, Integer)) As Integer
        For i As Integer = -1 To 1
            If field.V1 + i >= 0 AndAlso field.V1 + i < Me.m_width Then
                For j As Integer = -1 To 1
                    If field.V2 + j >= 0 AndAlso field.V2 + j < Me.m_height Then
                        If ((Me.m_field(field.V1 + i, field.V2 + j).State And FieldState.Flagged) = _
                            FieldState.Flagged _
                        ) Then
                            CountMinesAroundField += 1
                        End If
                    End If
                Next
            End If
        Next

        Return CountMinesAroundField
    End Function

    Private Sub ChangeFieldState(ByVal field As Field)
        If (field.State And FieldState.Open) = FieldState.Open Then
            Return
        ElseIf (field.State And FieldState.Flagged) = FieldState.Flagged Then
            field.State = field.State And (Not FieldState.Flagged)
            If Me.m_useMarker Then field.State = field.State Or FieldState.Unknown
            Me.m_minesLeft += 1
        ElseIf (field.State And FieldState.Unknown) = FieldState.Unknown Then
            field.State = field.State And (Not FieldState.Unknown)
        Else
            field.State = field.State Or FieldState.Flagged
            Me.m_minesLeft -= 1
        End If

        field.DoGraphicalUpdate = True
    End Sub

    Private Sub Redraw()
        Dim width As Integer = Me.m_uiDevice.Width - 5
        Dim height As Integer = Me.m_uiDevice.Height - 5

        Using g As Drawing.Graphics = Drawing.Graphics.FromImage(Me.m_bitmap)
            If Me.m_gameState = GameState.Paused Then
                g.Clear(Drawing.Color.LightGray)
                Dim pause As Drawing.Bitmap = My.Resources.pause
                g.DrawImage(pause, width \ 2 - pause.Width \ 2, height \ 2 - pause.Height \ 2, pause.Width, pause.Height)
                Me.m_uiDevice.Refresh()
                Return
            End If

            For x As Integer = 0 To Me.m_width - 1
                For y As Integer = 0 To Me.m_height - 1
                    If Me.m_field(x, y).DoGraphicalUpdate Then
                        If (Me.m_field(x, y).State And FieldState.Open) = FieldState.Open Then
                            g.FillRectangle(Drawing.Brushes.LightGray, x * 16, y * 16, 16, 16)
                            g.DrawLine(Drawing.Pens.DarkGray, x * 16, y * 16, x * 16 + 15, y * 16)
                            g.DrawLine(Drawing.Pens.DarkGray, x * 16, y * 16, x * 16, y * 16 + 15)

                            Select Case Me.m_field(x, y).Value
                                Case FieldValue.Mined
                                    If (Me.m_field(x, y).State And FieldState.ClickedBomb) = FieldState.ClickedBomb Then
                                        g.FillRectangle(Drawing.Brushes.Red, x * 16, y * 16, 16, 16)
                                    End If

                                    g.DrawImage(My.Resources.bomb, x * 16 + 2, y * 16 + 2, 13, 13)
                                Case FieldValue.Valued
                                    If Me.m_field(x, y).Number > 0 Then
                                        g.DrawImage(GetNumberPic(Me.m_field(x, y).Number), x * 16 + 2, y * 16 + 2, 13, 13)
                                    End If
                            End Select
                        Else
                            g.DrawLine(Drawing.Pens.White, x * 16, y * 16, x * 16, y * 16 + 15)
                            g.DrawLine(Drawing.Pens.White, x * 16, y * 16, x * 16 + 15, y * 16)
                            g.DrawLine(Drawing.Pens.White, x * 16 + 1, y * 16 + 1, x * 16 + 1, y * 16 + 14)
                            g.DrawLine(Drawing.Pens.White, x * 16 + 1, y * 16 + 1, x * 16 + 14, y * 16 + 1)

                            g.DrawLine(Drawing.Pens.DarkGray, x * 16 + 15, y * 16, x * 16 + 15, y * 16 + 15)
                            g.DrawLine(Drawing.Pens.DarkGray, x * 16, y * 16 + 15, x * 16 + 15, y * 16 + 15)
                            g.DrawLine(Drawing.Pens.DarkGray, x * 16 + 14, y * 16 + 1, x * 16 + 14, y * 16 + 14)
                            g.DrawLine(Drawing.Pens.DarkGray, x * 16 + 1, y * 16 + 14, x * 16 + 14, y * 16 + 14)
                            g.FillRectangle(Drawing.Brushes.LightGray, x * 16 + 2, y * 16 + 2, 12, 12)

                            If (Me.m_field(x, y).State And FieldState.Flagged) = FieldState.Flagged Then
                                g.DrawImage(My.Resources.flag, x * 16 + 2, y * 16 + 2, 13, 13)
                            ElseIf (Me.m_field(x, y).State And FieldState.Unknown) = FieldState.Unknown Then
                                g.DrawImage(My.Resources.hidden, x * 16 + 2, y * 16 + 2, 13, 13)
                            ElseIf (Me.m_field(x, y).State And FieldState.WrongBomb) = FieldState.WrongBomb Then
                                g.DrawImage(My.Resources.wrongbomb, x * 16 + 2, y * 16 + 2, 13, 13)
                            End If
                        End If

                        Me.m_field(x, y).DoGraphicalUpdate = False
                    End If
                Next
            Next

            If Me.m_activeFields IsNot Nothing Then
                For Each field As Pair(Of Integer, Integer) In Me.m_activeFields
                    With field
                        If Not (Me.m_field(.V1, .V2).State And FieldState.Open) = FieldState.Open Then
                            g.FillRectangle(Drawing.Brushes.LightGray, .V1 * 16, .V2 * 16, 16, 16)
                            g.DrawLine(Drawing.Pens.DarkGray, .V1 * 16, .V2 * 16, .V1 * 16 + 15, .V2 * 16)
                            g.DrawLine(Drawing.Pens.DarkGray, .V1 * 16, .V2 * 16, .V1 * 16, .V2 * 16 + 15)

                            If (Me.m_field(.V1, .V2).State And FieldState.Unknown) = FieldState.Unknown Then
                                g.DrawImage(My.Resources.hidden, .V1 * 16 + 2, .V2 * 16 + 2, 13, 13)
                            End If
                        End If
                    End With
                Next
            End If
        End Using

        Me.m_uiDevice.Refresh()
    End Sub

    Private Sub InitGraphics()
        Me.m_bitmap = New Drawing.Bitmap(Me.m_uiDevice.Width, Me.m_uiDevice.Height)
    End Sub

    Private Function GetNumberPic(ByVal n As Integer) As Drawing.Bitmap
        Return CType(My.Resources.ResourceManager.GetObject("n" & n.ToString()), Drawing.Bitmap)
    End Function

    Private Sub GameIsWon()
        Me.m_timer.Stop()
        Me.m_gameState = GameState.Won

        For Each field As Field In Me.m_field
            If field.Value = FieldValue.Mined Then
                field.State = FieldState.Flagged
                field.DoGraphicalUpdate = True
            End If
        Next
    End Sub

    Private Sub GameIsLost()
        Me.m_timer.Stop()
        Me.m_gameState = GameState.Lost

        For Each field As Field In Me.m_field
            If field.Value = FieldValue.Mined Then
                field.State = field.State Or FieldState.Open
                field.DoGraphicalUpdate = True
            Else
                If (field.State And FieldState.Flagged) = FieldState.Flagged Then
                    field.State = field.State And Not FieldState.Flagged
                    field.State = field.State Or FieldState.WrongBomb
                    field.DoGraphicalUpdate = True
                End If
            End If
        Next
    End Sub

#End Region

End Class